package com.faraphel.tasks_valider.connectivity.wifi_p2p

import android.content.BroadcastReceiver
import android.content.Context
import android.content.Intent
import android.content.IntentFilter
import android.net.wifi.p2p.WifiP2pConfig
import android.net.wifi.p2p.WifiP2pDeviceList
import android.net.wifi.p2p.WifiP2pManager
import android.os.Build
import android.util.Log
import androidx.annotation.RequiresApi
import com.faraphel.tasks_valider.connectivity.wifi_p2p.error.WifiP2pHelperInvalidActionException


/**
 * A helper to wrap the WiFi-Direct manager, the channel and the events.
 *
 * This avoids certain annoying features such as always specifying the channel as the first argument or
 * handling all the events with the base event system.
 *
 * @param manager: the WiFi-Direct base manager
 * @param channel: the WiFi-Direct channel
 */
class WifiP2pHelper(
    private val manager: WifiP2pManager,
    private val channel: WifiP2pManager.Channel,
) : BroadcastReceiver() {
    companion object {
        /**
         * This IntentFilter can be used to subscribe the helper to every supported WiFi-Direct events.
         */
        val ALL_INTENT_FILTER = IntentFilter().apply {
            addAction(WifiP2pManager.WIFI_P2P_STATE_CHANGED_ACTION)
            addAction(WifiP2pManager.WIFI_P2P_PEERS_CHANGED_ACTION)
            addAction(WifiP2pManager.WIFI_P2P_CONNECTION_CHANGED_ACTION)
            addAction(WifiP2pManager.WIFI_P2P_THIS_DEVICE_CHANGED_ACTION)
        }
    }

    // Events

    /**
     * Contains the listeners to call when the peers are updated.
     */
    private val listenersPeersChanged = mutableListOf<WifiP2pHelperListenerInfo<(WifiP2pDeviceList?) -> Unit>>()

    /**
     * Event receiver for the WiFi-Direct events.
     * Should not be called manually.
     * @param context: the WiFi-Direct context
     * @param intent: the event information
     */
    @RequiresApi(Build.VERSION_CODES.N)
    override fun onReceive(context: Context, intent: Intent) {
        when (intent.action) {
            WifiP2pManager.WIFI_P2P_CONNECTION_CHANGED_ACTION -> {
                Log.v("wifi-p2p", "Connection Changed : $context, $intent, ${intent.extras}")
            }

            WifiP2pManager.WIFI_P2P_PEERS_CHANGED_ACTION -> {
                Log.v("wifi-p2p", "Peers Changed : $context, $intent, ${intent.extras}")

                this.requestPeers { peers ->
                    this@WifiP2pHelper.listenersPeersChanged.forEach { listenerInfo -> listenerInfo.callback(peers) }
                    this@WifiP2pHelper.listenersPeersChanged.removeIf { listenerInfo -> listenerInfo.once }
                }
            }

            WifiP2pManager.WIFI_P2P_STATE_CHANGED_ACTION -> {
                Log.v("wifi-p2p", "State Changed : $context, $intent, ${intent.extras}")
            }

            WifiP2pManager.WIFI_P2P_THIS_DEVICE_CHANGED_ACTION -> {
                Log.v("wifi-p2p", "This Device Changed : $context, $intent, ${intent.extras}")
            }
        }
    }

    /**
     * Register a listener to a specific WiFi-Direct event.
     * @param action: the name of the WiFi-Direct event
     * @param callback: the callback function
     * @param once: should the event be only called once ?
     */
    fun <CallbackType> registerListener(action: String, callback: CallbackType, once: Boolean = false) {
        when (action) {
            WifiP2pManager.WIFI_P2P_PEERS_CHANGED_ACTION -> {
                @Suppress("UNCHECKED_CAST")
                this.listenersPeersChanged.add(WifiP2pHelperListenerInfo(
                    callback as ((WifiP2pDeviceList?) -> Unit),
                    once
                ))
            }
            else -> {
                // raise an exception if the action is invalid
                throw WifiP2pHelperInvalidActionException(action)
            }
        }
    }

    @RequiresApi(Build.VERSION_CODES.N)
    fun <CallbackType> unregisterListener(action: String, callback: CallbackType) {
        when (action) {
            WifiP2pManager.WIFI_P2P_PEERS_CHANGED_ACTION -> {
                this.listenersPeersChanged.removeIf { listenerInfo -> listenerInfo.callback == callback }
            }
            else -> {
                // raise an exception if the action is invalid
                throw WifiP2pHelperInvalidActionException(action)
            }
        }
    }

    // Wrappers

    /**
     * Wrapper around WifiP2pManager::connect
     *
     * Connect to another device, allowing for a communication using Sockets
     */
    fun connect(config: WifiP2pConfig, listener: WifiP2pManager.ActionListener? = null) {
        this.manager.connect(this.channel, config, listener)
    }

    /**
     * Wrapper around WifiP2pManager::requestPeers
     *
     * Request the list of peers after a discovery.
     */
    fun requestPeers(listener: WifiP2pManager.PeerListListener? = null) {
        this.manager.requestPeers(this.channel, listener)
    }

    /**
     * Wrapper around WifiP2pManager::discoverPeers
     *
     * Start discovering peers.
     * Once founds, the WIFI_P2P_PEERS_CHANGED_ACTION event will be triggered.
     */
    fun discoverPeers(listener: WifiP2pManager.ActionListener? = null) {
        this.manager.discoverPeers(this.channel, listener)
    }

    /**
     * Wrapper around WifiP2pManager::requestConnectionInfo
     *
     * Obtain information about a connection with another device.
     */
    fun requestConnectionInfo(listener: WifiP2pManager.ConnectionInfoListener? = null) {
        this.manager.requestConnectionInfo(this.channel, listener)
    }

    /**
     * Wrapper around WifiP2pManager::requestGroupInfo
     *
     * Obtain information about the current group.
     */
    fun requestGroupInfo(listener: WifiP2pManager.GroupInfoListener? = null) {
        this.manager.requestGroupInfo(this.channel, listener)
    }

    /**
     * Wrapper around WifiP2pManager::createGroup
     *
     * Create a new WiFi-Direct group.
     */
    fun createGroup(listener: WifiP2pManager.ActionListener? = null) {
        this.manager.createGroup(this.channel, listener)
    }

    /**
     * Wrapper around WifiP2pManager::removeGroup
     *
     * Disconnect from the current WiFi-Direct group.
     */
    fun removeGroup(listener: WifiP2pManager.ActionListener? = null) {
        this.manager.removeGroup(this.channel, listener)
    }

    /**
     * Create a new WiFi-Direct group.
     * If already connected to a group, quit it first.
     *
     * @param listener: the createGroup listener
     * @param onFailure: error handler for the removeGroup event
     */
    fun recreateGroup(
        listener: WifiP2pManager.ActionListener? = null,
        onFailure: ((Int) -> Unit)? = null
    ) {
        // get the current group information
        this.requestGroupInfo { group ->
            // if a group exist, quit it
            if (group != null)
                this.removeGroup(object : WifiP2pManager.ActionListener {
                    override fun onSuccess() {
                        // once left, create the group
                        this@WifiP2pHelper.createGroup(listener)
                    }

                    override fun onFailure(reason: Int) {
                        // call the handler
                        onFailure?.invoke(reason)
                    }

                })
            else
                // create the group
                this.createGroup(listener)
        }
    }
}